
/* Calculates dl and Cl based on formuli given in */
/* Snodgrass and Corwin (1988) JEP:GEN, Vol. 117, No. 1, 34-50 */
/* p value */
/* Input from stdin, two columns of numbers */

#include <stdio.h>

extern double log();

#define PTS 1000
#define FALSE 0
#define TRUE  1

int 	Maxlines = 1000;

main (argc, argv)
int argc;
char *argv[];
{

  int 	nlines;
  char	**lptr;
  register	int 	i;

  char line[80];
  double h_in, fa_in;
  double sig_n, noi_n;
  double H, FA;

  double dl, Cl;

  if (argc > 1 )
    if (argv[1][1] == 'h') {
      printf ("Computes sensitivity and bias using a signal detection\n");
      printf ("model with logistic distributions.\n");
      printf ("usage: [hits n_signal fa's n_noise] from stdin.\n");
      printf ("output: [corrected_hits corrected_fa's dl Cl] to stdout.\n");
      printf ("Reference:\n");
      printf ("Snodgrass, J. G. & Corwin, J. (1988) JEP:GEN, Vol. 117, No. 1, 34-50.\n");
      exit (0);
    }

  nlines = readlines (&lptr, Maxlines, stdin);

  for (i = 0; i < nlines; i++)
    {

    sscanf(lptr[i], "%lf %lf %lf %lf", &h_in, &sig_n, &fa_in, &noi_n);

    H = (h_in + 0.5) / (sig_n + 1.0);
    FA = (fa_in + 0.5) / (noi_n + 1.0);

    dl = log( (H * (1.0 - FA)) / (FA * (1.0 - H)) );

    Cl = 0.5 * log( ((1.0 - FA) * (1 - H)) / (H * FA) );

    printf("%4.3f\t%4.3f\t%4.3f\t%4.3f\n", H, FA, dl, Cl);
    }
}

