
/* Calculates independent samples t statistic and associated two-tailed */
/* p value */
/* Input from stdin, two columns of numbers */

#include <stdio.h>

extern double sqrt();

#define PTS 1000
#define FALSE 0
#define TRUE  1

int 	Maxlines = 1000;

/*		Copyright (c) 1981 Gary Perlman
This software may be copied freely provided:  (1) it is not used for
personal or material gain, and (2) this notice accompanies each copy.

Disclaimer:  No gauarantees of performance accompany this software,
nor is any responsbility assumed on the part of the author.  All the
software has been tested extensively and every effort has been made to
insure its reliability.
This notice applies only to pot a modified version of pof.c */
double atan();

double pot (t, n) double t;
{
	int i, j, a, b, m;
	double	w, y, z, d, p;

	m = 1; /* since this is a t value df(effect) = 1 */

	if (n <= 0) return (1.0);

	a = m%2 ? 1 : 2;
	b = n%2 ? 1 : 2;
	w = t * t * m / n;
	z = 1.0 / (1.0 + w);
	if (a == 1)
	  if (b == 1) {
		p = sqrt (w);
		y = 0.3183098862;
		d = y * z / p;
		p = 2 * y * atan (p);
	}
	else {
		p = sqrt (w * z);
		d = 0.5 * p * z / w;
	}
	else if (b == 1) {
		p = sqrt (z);
		d = 0.5 * z * p;
		p = 1.0 - p;
	}
	else {
		d = z * z;
		p = w * z;
	}
	y = 2.0 * w / z;
	for (j = b + 2; j <= n; j += 2)	{
		d *= (1.0 + a / (j - 2.0)) * z;
		p = (a == 1 ? p + d * y / (j - 1.0) : (p + w) * z);
	}
	y = w * z;
	z = 2.0 / z;
	b = n - 2;
	for (i = a + 2; i <= m; i += 2) {
		j = i + b;
		d *= y * j / (i - 2.0);
		p -= z * d / j;
	}
	return ( (1.0-p) );
}

/* Formula's for calculation taken from Rosenthal and Rosnow, 1984, */
/* "Essentials of Behavioral Research", page 232 */

main()
{

  int 	nlines;
  char	**lptr;
  register	int 	ii;

  char line[80];
  float xin, yin;
  float x[PTS], y[PTS];
  int n, df;
  float xsum, ysum;
  float xmn, ymn;
  float xam, yam;
  float var;
  float t, p;
  int i, j;

  n = 0;
  xsum = 0.0;
  ysum = 0.0;
  xam = 0.0;
  yam = 0.0;

  nlines = readlines (&lptr, Maxlines, stdin);

  for (ii = 0; ii < nlines; ii++)
    {
      sscanf(lptr[ii], "%f %f", &xin, &yin);
      x[n] = xin;
      y[n] = yin;
      xsum += xin;
      ysum += yin;
      ++n;
    }

	if (n == 0) {
		printf("No data points!\n");
		exit();
	}

	if (n == 1) {
		printf("Not enough data points!\n");
		exit();
	}

	df = n + n - 2;
	xmn = xsum / n;
	ymn = ysum / n;

	for (i = 0; i < n; i++) {
		xam += (x[i] - xmn) * (x[i] - xmn);
		yam += (y[i] - ymn) * (y[i] - ymn);
	}

	var = (xam + yam) / df;

	t = (xmn - ymn) / sqrt( (2.0 / n) * var );

	p = pot(t, df);

	printf("t(%d) = %7.3f, p = %4.3f (two-tailed)\n", df, t, p);
}
